import os

import pytest
from pymetastore.hive_metastore import ttypes
from pymetastore.hive_metastore.ThriftHiveMetastore import Client
from pymetastore.metastore import HMS
from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket, TTransport

from recap.clients import create_client
from recap.clients.hive_metastore import HiveMetastoreClient
from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


@pytest.fixture(scope="module")
def hive_client():
    host = os.environ.get("HMS_HOST", "localhost")
    port = int(os.environ.get("HMS_PORT", 9083))
    transport = TSocket.TSocket(host, port)
    transport = TTransport.TBufferedTransport(transport)
    protocol = TBinaryProtocol.TBinaryProtocol(transport)
    client = Client(protocol)
    transport.open()

    yield client

    transport.close()


@pytest.fixture(scope="module", autouse=True)
def setup_data(hive_client):
    db = ttypes.Database(
        name="test_db",
        description="This is a test database",
        locationUri="/tmp/test_db.db",
        parameters={},
        ownerName="owner",
    )

    if "test_db" in hive_client.get_all_databases():
        hive_client.drop_database("test_db", True, True)
    hive_client.create_database(db)

    partitionKeys = [ttypes.FieldSchema(name="partition", type="string", comment="")]

    serde_info = ttypes.SerDeInfo(
        name="test_serde",
        serializationLib="org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe",
        parameters={"field.delim": ","},
    )

    # testing primitive Types
    cols2 = [
        ttypes.FieldSchema(name="col4", type="boolean", comment="c4"),
        ttypes.FieldSchema(name="col5", type="tinyint", comment="c5"),
        ttypes.FieldSchema(name="col6", type="smallint", comment="c6"),
        ttypes.FieldSchema(name="col7", type="bigint", comment="c7"),
        ttypes.FieldSchema(name="col8", type="float", comment="c8"),
        ttypes.FieldSchema(name="col9", type="double", comment="c9"),
        ttypes.FieldSchema(name="col10", type="date", comment="c10"),
        ttypes.FieldSchema(name="col11", type="timestamp", comment="c11"),
        ttypes.FieldSchema(
            name="col12",
            type="timestamp with local time zone",
            comment="c12",
        ),
        ttypes.FieldSchema(name="col13", type="interval_year_month", comment="c13"),
        ttypes.FieldSchema(name="col14", type="interval_day_time", comment="c14"),
        ttypes.FieldSchema(name="col15", type="binary", comment="c15"),
        ttypes.FieldSchema(name="col16", type="void", comment="c16"),
    ]

    storageDesc = ttypes.StorageDescriptor(
        location="/tmp/test_db/test_table2",
        cols=cols2,
        inputFormat="org.apache.hadoop.mapred.TextInputFormat",
        outputFormat="org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
        compressed=False,
        numBuckets=-1,
        serdeInfo=serde_info,
        bucketCols=[],
    )

    table = ttypes.Table(
        tableName="test_table2",
        dbName="test_db",
        owner="owner",
        createTime=0,
        lastAccessTime=0,
        retention=0,
        sd=storageDesc,
        partitionKeys=partitionKeys,
        parameters={},
        tableType="EXTERNAL_TABLE",
    )

    if "test_table2" in hive_client.get_all_tables("test_db"):
        hive_client.drop_table("test_db", "test_table2", True)
    hive_client.create_table(table)

    # testing Parameterized Types
    cols3 = [
        ttypes.FieldSchema(name="col16", type="decimal(10,2)", comment="c16"),
        ttypes.FieldSchema(name="col17", type="varchar(10)", comment="c17"),
        ttypes.FieldSchema(name="col18", type="char(10)", comment="c18"),
        ttypes.FieldSchema(name="col19", type="array<int>", comment="c19"),
        ttypes.FieldSchema(name="col20", type="map<int,string>", comment="c20"),
        ttypes.FieldSchema(name="col21", type="struct<a:int,b:string>", comment="c21"),
        ttypes.FieldSchema(name="col22", type="uniontype<int,string>", comment="c22"),
    ]

    storageDesc = ttypes.StorageDescriptor(
        location="/tmp/test_db/test_table3",
        cols=cols3,
        inputFormat="org.apache.hadoop.mapred.TextInputFormat",
        outputFormat="org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
        compressed=False,
        numBuckets=-1,
        serdeInfo=serde_info,
        bucketCols=[],
    )

    table = ttypes.Table(
        tableName="test_table3",
        dbName="test_db",
        owner="owner",
        createTime=0,
        lastAccessTime=0,
        retention=0,
        sd=storageDesc,
        partitionKeys=partitionKeys,
        parameters={},
        tableType="EXTERNAL_TABLE",
    )

    if "test_table3" in hive_client.get_all_tables("test_db"):
        hive_client.drop_table("test_db", "test_table3", True)
    hive_client.create_table(table)


def test_primitive_types(hive_client):
    hms = HMS(hive_client)
    client = HiveMetastoreClient(hms)
    table = client.get_schema("test_db", "test_table2")
    fields = table.fields

    assert len(fields) == 13

    assert fields[0].extra_attrs["name"] == "col4"
    assert isinstance(fields[0], UnionType)
    assert isinstance(fields[0].types[0], NullType)
    assert isinstance(fields[0].types[1], BoolType)
    assert fields[0].doc == "c4"

    assert fields[1].extra_attrs["name"] == "col5"
    assert isinstance(fields[1], UnionType)
    assert isinstance(fields[1].types[0], NullType)
    assert isinstance(fields[1].types[1], IntType)
    assert fields[1].types[1].bits == 8
    assert fields[1].types[1].signed is True
    assert fields[1].doc == "c5"

    assert fields[2].extra_attrs["name"] == "col6"
    assert isinstance(fields[2], UnionType)
    assert isinstance(fields[2].types[0], NullType)
    assert isinstance(fields[2].types[1], IntType)
    assert fields[2].types[1].bits == 16
    assert fields[2].types[1].signed is True
    assert fields[2].doc == "c6"

    assert fields[3].extra_attrs["name"] == "col7"
    assert isinstance(fields[3], UnionType)
    assert isinstance(fields[3].types[0], NullType)
    assert isinstance(fields[3].types[1], IntType)
    assert fields[3].types[1].bits == 64
    assert fields[3].types[1].signed is True
    assert fields[3].doc == "c7"

    assert fields[4].extra_attrs["name"] == "col8"
    assert isinstance(fields[4], UnionType)
    assert isinstance(fields[4].types[0], NullType)
    assert isinstance(fields[4].types[1], FloatType)
    assert fields[4].types[1].bits == 32
    assert fields[4].doc == "c8"

    assert fields[5].extra_attrs["name"] == "col9"
    assert isinstance(fields[5], UnionType)
    assert isinstance(fields[5].types[0], NullType)
    assert isinstance(fields[5].types[1], FloatType)
    assert fields[5].types[1].bits == 64
    assert fields[5].doc == "c9"

    assert fields[6].extra_attrs["name"] == "col10"
    assert isinstance(fields[6], UnionType)
    assert isinstance(fields[6].types[0], NullType)
    assert isinstance(fields[6].types[1], IntType)
    assert fields[6].types[1].logical == "build.recap.Date"
    assert fields[6].types[1].bits == 32
    assert fields[6].types[1].signed is True
    assert fields[6].types[1].extra_attrs["unit"] == "day"
    assert fields[6].doc == "c10"

    assert fields[7].extra_attrs["name"] == "col11"
    assert isinstance(fields[7], UnionType)
    assert isinstance(fields[7].types[0], NullType)
    assert isinstance(fields[7].types[1], IntType)
    assert fields[7].types[1].logical == "build.recap.Timestamp"
    assert fields[7].types[1].bits == 64
    assert fields[7].types[1].signed is True
    assert fields[7].types[1].extra_attrs["unit"] == "nanosecond"
    assert fields[7].types[1].extra_attrs["timezone"] == "UTC"
    assert fields[7].doc == "c11"

    assert fields[8].extra_attrs["name"] == "col12"
    assert isinstance(fields[8], UnionType)
    assert isinstance(fields[8].types[0], NullType)
    assert isinstance(fields[8].types[1], IntType)
    assert fields[8].types[1].logical == "build.recap.Timestamp"
    assert fields[8].types[1].bits == 64
    assert fields[8].types[1].signed is True
    assert fields[8].types[1].extra_attrs["unit"] == "nanosecond"
    assert fields[8].types[1].extra_attrs["timezone"] is None
    assert fields[8].doc == "c12"

    assert fields[9].extra_attrs["name"] == "col13"
    assert isinstance(fields[9], UnionType)
    assert isinstance(fields[9].types[0], NullType)
    assert isinstance(fields[9].types[1], BytesType)
    assert fields[9].doc == "c13"

    assert fields[10].extra_attrs["name"] == "col14"
    assert isinstance(fields[10], UnionType)
    assert isinstance(fields[10].types[0], NullType)
    assert isinstance(fields[10].types[1], BytesType)
    assert fields[10].doc == "c14"

    assert fields[11].extra_attrs["name"] == "col15"
    assert isinstance(fields[11], UnionType)
    assert isinstance(fields[11].types[0], NullType)
    assert isinstance(fields[11].types[1], BytesType)
    assert fields[11].types[1].bytes_ == 2_147_483_647
    assert fields[11].doc == "c15"

    assert fields[12].extra_attrs["name"] == "col16"
    assert isinstance(fields[12], NullType)
    assert fields[12].doc == "c16"


def test_parameterized_types(hive_client):
    hms = HMS(hive_client)
    client = HiveMetastoreClient(hms)
    table = client.get_schema("test_db", "test_table3")
    fields = table.fields

    assert len(fields) == 7

    assert fields[0].extra_attrs["name"] == "col16"
    assert isinstance(fields[0], UnionType)
    assert isinstance(fields[0].types[0], NullType)
    assert isinstance(fields[0].types[1], BytesType)
    assert fields[0].types[1].logical == "build.recap.Decimal"
    assert fields[0].types[1].bytes_ == 16
    assert fields[0].types[1].extra_attrs["precision"] == 10
    assert fields[0].types[1].extra_attrs["scale"] == 2
    assert fields[0].doc == "c16"

    assert fields[1].extra_attrs["name"] == "col17"
    assert isinstance(fields[1], UnionType)
    assert isinstance(fields[1].types[0], NullType)
    assert isinstance(fields[1].types[1], StringType)
    assert fields[1].types[1].bytes_ == 10
    assert fields[1].doc == "c17"

    assert fields[2].extra_attrs["name"] == "col18"
    assert isinstance(fields[2], UnionType)
    assert isinstance(fields[2].types[0], NullType)
    assert isinstance(fields[2].types[1], StringType)
    assert fields[2].types[1].bytes_ == 10
    assert fields[2].doc == "c18"

    assert fields[3].extra_attrs["name"] == "col19"
    assert isinstance(fields[3], UnionType)
    assert isinstance(fields[3].types[0], NullType)
    assert isinstance(fields[3].types[1], ListType)
    assert isinstance(fields[3].types[1].values, UnionType)
    assert isinstance(fields[3].types[1].values.types[0], NullType)
    assert isinstance(fields[3].types[1].values.types[1], IntType)
    assert fields[3].types[1].values.types[1].bits == 32
    assert fields[3].types[1].values.types[1].signed
    assert fields[3].doc == "c19"

    assert fields[4].extra_attrs["name"] == "col20"
    assert isinstance(fields[4], UnionType)
    assert isinstance(fields[4].types[0], NullType)
    assert isinstance(fields[4].types[1], MapType)
    assert isinstance(fields[4].types[1].keys, UnionType)
    assert isinstance(fields[4].types[1].keys.types[0], NullType)
    assert isinstance(fields[4].types[1].keys.types[1], IntType)
    assert fields[4].types[1].keys.types[1].bits == 32
    assert fields[4].types[1].keys.types[1].signed
    assert isinstance(fields[4].types[1].values, UnionType)
    assert isinstance(fields[4].types[1].values.types[0], NullType)
    assert isinstance(fields[4].types[1].values.types[1], StringType)
    assert fields[4].types[1].values.types[1].bytes_ == 9_223_372_036_854_775_807
    assert fields[4].doc == "c20"

    assert fields[5].extra_attrs["name"] == "col21"
    assert isinstance(fields[5], UnionType)
    assert isinstance(fields[5].types[0], NullType)
    assert isinstance(fields[5].types[1], StructType)
    assert len(fields[5].types[1].fields) == 2
    assert fields[5].types[1].fields[0].extra_attrs["name"] == "a"
    assert isinstance(fields[5].types[1].fields[0], UnionType)
    assert isinstance(fields[5].types[1].fields[0].types[0], NullType)
    assert isinstance(fields[5].types[1].fields[0].types[1], IntType)
    assert fields[5].types[1].fields[0].types[1].bits == 32
    assert fields[5].types[1].fields[0].types[1].signed
    assert fields[5].types[1].fields[1].extra_attrs["name"] == "b"
    assert isinstance(fields[5].types[1].fields[1], UnionType)
    assert isinstance(fields[5].types[1].fields[1].types[0], NullType)
    assert isinstance(fields[5].types[1].fields[1].types[1], StringType)
    assert fields[5].types[1].fields[1].types[1].bytes_ == 9_223_372_036_854_775_807
    assert fields[5].doc == "c21"

    assert fields[6].extra_attrs["name"] == "col22"
    assert isinstance(fields[6], UnionType)
    assert len(fields[6].types) == 3
    assert isinstance(fields[6].types[0], NullType)
    assert isinstance(fields[6].types[1], IntType)
    assert fields[6].types[1].bits == 32
    assert fields[6].types[1].signed
    assert isinstance(fields[6].types[2], StringType)
    assert fields[6].types[2].bytes_ == 9_223_372_036_854_775_807
    assert fields[6].doc == "c22"


def test_create_client():
    hms_url = "thrift+hms://hive:password@localhost:9083/test_db"

    with create_client(hms_url) as client:
        assert client.ls() == ["default", "test_db"]
        assert client.ls("test_db") == ["test_table2", "test_table3"]



import threading
import time

import httpx
import psycopg2
from uvicorn import Server
from uvicorn.config import Config

from recap.gateway import app

client = httpx.Client(follow_redirects=True, base_url="http://localhost:8000")


class TestGateway:
    server = None
    server_thread = None

    @classmethod
    def setup_class(cls):
        # Connect to the PostgreSQL database
        cls.connection = psycopg2.connect(
            host="localhost",
            port="5432",
            user="postgres",
            password="password",
            dbname="testdb",
        )

        # Create tables
        cursor = cls.connection.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS test_types (test_integer INTEGER);")
        cls.connection.commit()

        # Start FastAPI app
        config = Config(app=app, host="127.0.0.1", port=8000)
        cls.server = Server(config=config)
        cls.server_thread = threading.Thread(target=cls.server.run)
        cls.server_thread.start()
        cls.server.server_state

        # Wait for the server to start
        checks = 0
        while cls.server.started is False and checks < 10:
            time.sleep(1)
            checks += 1
        if cls.server.started is False:
            raise RuntimeError("Unable to start gateway server after 10 seconds")

    @classmethod
    def teardown_class(cls):
        # Delete tables
        cursor = cls.connection.cursor()
        cursor.execute("DROP TABLE IF EXISTS test_types;")
        cls.connection.commit()

        # Close the connection
        cls.connection.close()

        # Stop the FastAPI app
        if server := cls.server:
            server.should_exit = True
        if server_thread := cls.server_thread:
            server_thread.join(10)
            if server_thread.is_alive():
                raise RuntimeError("Unable to shutdown gateway server after 10 seconds")

    def test_ls_root(self):
        response = client.get("/ls")
        assert response.status_code == 200
        assert response.json() == ["pg"]

    def test_ls_subpath(self):
        response = client.get("/ls/pg")
        assert response.status_code == 200
        assert response.json() == ["postgres", "template0", "template1", "testdb"]

    def test_schema(self):
        response = client.get("/schema/pg/testdb/public/test_types")
        assert response.status_code == 200
        assert response.json() == {
            "type": "struct",
            "fields": [{"name": "test_integer", "type": "int32", "optional": True}],
        }



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from google.cloud import bigquery

from recap.converters.bigquery import BigQueryConverter
from recap.types import StructType


class BigQueryClient:
    def __init__(self, client: bigquery.Client):
        self.client = client

    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[BigQueryClient, None, None]:
        with bigquery.Client() as client:
            yield BigQueryClient(client)

    def ls(self, project: str | None = None, dataset: str | None = None) -> list[str]:
        match (project, dataset):
            case (None, None):
                return self.ls_projects()
            case (str(project), None):
                return self.ls_datasets(project)
            case (str(project), str(dataset)):
                return self.ls_tables(project, dataset)
            case _:
                raise ValueError("Invalid arguments")

    def ls_projects(self) -> list[str]:
        return [project.project_id for project in self.client.list_projects()]

    def ls_datasets(self, project: str) -> list[str]:
        return [dataset.dataset_id for dataset in self.client.list_datasets(project)]

    def ls_tables(self, project: str, dataset: str) -> list[str]:
        dataset_ref = self.client.dataset(dataset, project)
        return [table.table_id for table in self.client.list_tables(dataset_ref)]

    def get_schema(self, project: str, dataset: str, table: str, **_) -> StructType:
        table_ref = self.client.dataset(dataset, project).table(table)
        table_obj = self.client.get_table(table_ref)
        return BigQueryConverter().to_recap(table_obj.schema)



from __future__ import annotations

from abc import ABC
from typing import Any, List, Protocol, Tuple

from recap.types import StructType


class DbapiClient(ABC):
    def __init__(self, connection: Connection, converter: DbapiConverter) -> None:
        self.connection = connection
        self.converter = converter

    def ls(self, catalog: str | None = None, schema: str | None = None) -> list[str]:
        match (catalog, schema):
            case (None, None):
                return self.ls_catalogs()
            case (str(catalog), None):
                return self.ls_schemas(catalog)
            case (str(catalog), str(schema)):
                return self.ls_tables(catalog, schema)
            case _:
                raise ValueError("Invalid arguments")

    def ls_catalogs(self) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute(
            """
                SELECT catalog_name
                FROM information_schema.catalogs
                ORDER BY catalog_name ASC
            """
        )
        return [row[0] for row in cursor.fetchall()]

    def ls_schemas(self, catalog: str) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
                SELECT schema_name
                FROM information_schema.schemata
                WHERE catalog_name = {self.param_style}
            """,
            (catalog,),
        )
        return [row[0] for row in cursor.fetchall()]

    def ls_tables(self, catalog: str, schema: str) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
                SELECT table_name
                FROM information_schema.tables
                WHERE table_catalog = {self.param_style}
                    AND table_schema = {self.param_style}
            """,
            (catalog, schema),
        )
        return [row[0] for row in cursor.fetchall()]

    def get_schema(self, catalog: str, schema: str, table: str) -> StructType:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
                SELECT *
                FROM information_schema.columns
                WHERE table_name = {self.param_style}
                    AND table_schema = {self.param_style}
                    AND table_catalog = {self.param_style}
                ORDER BY ordinal_position ASC
            """,
            (table, schema, catalog),
        )

        names = [name[0].upper() for name in cursor.description]
        return self.converter.to_recap(
            # Make each row be a dict with the column names as keys
            [dict(zip(names, row)) for row in cursor.fetchall()]
        )

    @property
    def param_style(cls) -> str:
        return "%s"


class Connection(Protocol):
    def close(self) -> None:
        ...

    def commit(self) -> None:
        ...

    def rollback(self) -> None:
        ...

    def cursor(self) -> Cursor:
        ...


class Cursor(Protocol):
    def execute(self, query: str, parameters: Tuple = ()) -> None:
        ...

    def executemany(self, query: str, parameter_list: List[Tuple]) -> None:
        ...

    def fetchone(self) -> Tuple:
        ...

    def fetchall(self) -> List[Tuple]:
        ...

    def fetchmany(self, size: int) -> List[Tuple]:
        ...

    @property
    def description(self) -> List[Tuple]:
        ...


class DbapiConverter(Protocol):
    def to_recap(self, columns: list[dict[str, Any]]) -> StructType:
        ...



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from confluent_kafka.schema_registry import SchemaRegistryClient

from recap.converters.avro import AvroConverter
from recap.converters.json_schema import JSONSchemaConverter
from recap.converters.protobuf import ProtobufConverter
from recap.types import StructType

ALLOWED_ATTRS = {
    "url",
    "ssl.ca.location",
    "ssl.key.location",
    "ssl.certificate.location",
    "basic.auth.user.info",
}


class ConfluentRegistryClient:
    def __init__(self, registry: SchemaRegistryClient):
        self.registry = registry

    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[ConfluentRegistryClient, None, None]:
        # Filter out kwargs that are not allowed by the SchemaRegistryClient
        kwargs = {k: v for k, v in kwargs.items() if k in ALLOWED_ATTRS}

        with SchemaRegistryClient(kwargs) as registry:
            yield ConfluentRegistryClient(registry)

    def ls(self) -> list[str]:
        return self.registry.get_subjects()

    def get_schema(self, subject: str) -> StructType:
        has_kv = subject.endswith("-key") or subject.endswith("-value")
        subject = subject if has_kv else f"{subject}-value"
        registered_schema = self.registry.get_latest_version(subject)
        schema_str = registered_schema.schema.schema_str
        match registered_schema.schema.schema_type:
            case "AVRO":
                return AvroConverter().to_recap(schema_str)
            case "JSON":
                return JSONSchemaConverter().to_recap(schema_str)
            case "PROTOBUF":
                return ProtobufConverter().to_recap(schema_str)
            case _:
                raise ValueError(
                    f"Unsupported schema type {registered_schema.schema.schema_type}"
                )



from recap.clients import _parse_url


def test_parse_url_basic():
    url = "http+csr://username:password@hostname:8080/path/to/resource?query=value"
    result = _parse_url(url)
    expected = {
        "scheme": "http+csr",
        "netloc": "username:password@hostname:8080",
        "path": "/path/to/resource",
        "params": "",
        "query": "query=value",
        "fragment": "",
        "username": "username",
        "password": "password",
        "hostname": "hostname",
        "port": 8080,
        "dialect": "http",
        "driver": "csr",
        "user": "username",
        "host": "hostname",
        "paths": ["path", "to", "resource"],
        "url": "http://username:password@hostname:8080/path/to/resource?query=value",
        "query": "value",
    }
    assert result == expected


def test_parse_url_no_plus():
    url = "https://hostname/path"
    result = _parse_url(url)
    assert result["scheme"] == "https"
    assert "dialect" not in result
    assert "driver" not in result


def test_parse_url_multiple_query_params():
    url = "http://hostname/path?param1=value1&param2=value2&param2=value3"
    result = _parse_url(url)
    assert result["param1"] == "value1"
    assert result["param2"] == ["value2", "value3"]



import os
from pathlib import Path

from dotenv import load_dotenv, set_key, unset_key
from pydantic import AnyUrl, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

load_dotenv()

CONFIG_FILE = os.environ.get("RECAP_CONFIG") or os.path.expanduser("~/.recap/config")
SECRETS_DIR = os.environ.get("RECAP_SECRETS")


def touch_config():
    config_path = Path(CONFIG_FILE)
    config_path.parent.mkdir(parents=True, exist_ok=True)
    config_path.touch(mode=0o600, exist_ok=True)
    if SECRETS_DIR:
        secrets_path = Path(SECRETS_DIR)
        secrets_path.mkdir(mode=0o700, parents=True, exist_ok=True)


touch_config()


class RecapSettings(BaseSettings):
    systems: dict[str, AnyUrl] = Field(default_factory=dict)
    model_config = SettingsConfigDict(
        # .env takes priority over CONFIG_FILE
        env_file=[CONFIG_FILE, ".env"],
        env_file_encoding="utf-8",
        env_prefix="recap_",
        env_nested_delimiter="__",
        secrets_dir=SECRETS_DIR,
    )


def set_config(key: str, val: str):
    set_key(CONFIG_FILE, key.upper(), val)


def unset_config(key: str):
    unset_key(CONFIG_FILE, key.upper())



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from recap.clients.dbapi import Connection, DbapiClient
from recap.converters.mysql import MysqlConverter

MYSQL_CATALOG_NAME = "def"


class MysqlClient(DbapiClient):
    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, MysqlConverter())

    @staticmethod
    @contextmanager
    def create(
        paths: list[str] | None = None, **kwargs
    ) -> Generator[MysqlClient, None, None]:
        import mysql.connector
        from mysql.connector.constants import DEFAULT_CONFIGURATION

        if paths:
            kwargs["database"] = paths[0]

        # Only include kwargs that are valid for the MySQL connector
        kwargs = {k: v for k, v in kwargs.items() if k in DEFAULT_CONFIGURATION}

        with mysql.connector.connect(**kwargs) as client:
            yield MysqlClient(client)  # type: ignore

    def ls(self, schema: str | None = None) -> list[str]:
        match schema:
            case None:
                return self.ls_schemas(MYSQL_CATALOG_NAME)
            case str(schema):
                return self.ls_tables(MYSQL_CATALOG_NAME, schema)
            case _:
                raise ValueError("Invalid arguments")



import os
import tempfile
from importlib import reload
from json import loads
from unittest.mock import patch

from pydantic import AnyUrl
from typer.testing import CliRunner

import recap.settings
from recap.cli import app
from recap.types import IntType, StructType

runner = CliRunner()


class TestCli:
    @classmethod
    def setup_class(cls):
        # Create a temporary .env file for the test
        cls.temp_file = tempfile.NamedTemporaryFile(delete=False)
        cls.temp_file_name = cls.temp_file.name

    @classmethod
    def teardown_class(cls):
        # Delete the temporary .env file
        os.unlink(cls.temp_file_name)

    @patch("recap.commands.ls")
    def test_ls_root(self, mock_ls):
        mock_ls.return_value = ["foo", "bar"]
        result = runner.invoke(app, ["ls"])
        assert result.exit_code == 0
        assert loads(result.stdout) == mock_ls.return_value

    @patch("recap.commands.ls")
    def test_ls_subpath(self, mock_ls):
        mock_ls.return_value = ["foo", "bar"]
        result = runner.invoke(app, ["ls", "/foo/bar"])
        assert result.exit_code == 0
        assert loads(result.stdout) == mock_ls.return_value

    @patch("recap.commands.schema")
    def test_schema(self, mock_schema):
        mock_schema.return_value = StructType([IntType(bits=32)])
        result = runner.invoke(app, ["schema", "foo"])
        assert result.exit_code == 0
        assert loads(result.stdout) == {"type": "struct", "fields": ["int32"]}

    def test_add_remove(self):
        with patch.dict(os.environ, {"RECAP_CONFIG": self.temp_file_name}, clear=False):
            # Reload after patching to reset CONFIG_FILE
            reload(recap.settings)
            from recap.settings import RecapSettings

            # Test set_config
            url = AnyUrl("scheme://user:pw@localhost:1234/db")
            result = runner.invoke(app, ["add", "test_system", url.unicode_string()])
            assert result.exit_code == 0
            assert result.stdout == ""
            assert RecapSettings().systems.get("test_system") == url

            # Test unset_config
            result = runner.invoke(app, ["remove", "test_system"])
            assert result.exit_code == 0
            assert result.stdout == ""
            assert RecapSettings().systems.get("test_system") is None



import os
import tempfile
from importlib import reload
from unittest.mock import patch

from pydantic import AnyUrl

import recap.settings


class TestSettings:
    @classmethod
    def setup_class(cls):
        # Create a temporary .env file for the test
        cls.temp_file = tempfile.NamedTemporaryFile(delete=False)
        cls.temp_file_name = cls.temp_file.name

    @classmethod
    def teardown_class(cls):
        # Delete the temporary .env file
        os.unlink(cls.temp_file_name)

    def test_set_and_unset_config(self):
        with patch.dict(os.environ, {"RECAP_CONFIG": self.temp_file_name}, clear=False):
            # Reload after patching to reset CONFIG_FILE
            reload(recap.settings)
            from recap.settings import RecapSettings, set_config, unset_config

            # Test set_config
            set_config("recap_systems__test_key", "foo://test_value")
            assert RecapSettings().systems.get("test_key") == AnyUrl("foo://test_value")

            # Test unset_config
            unset_config("recap_systems__test_key")
            assert RecapSettings().systems.get("test_key") is None



from fastapi import FastAPI, HTTPException

from recap import commands
from recap.types import to_dict

app = FastAPI()


@app.get("/ls/{path:path}")
async def ls(path: str = "/") -> list[str]:
    """
    List the children of a path.
    """

    children = commands.ls(path)
    if children is not None:
        return children
    raise HTTPException(status_code=404, detail="Path not found")


@app.get("/schema/{path:path}")
async def schema(path: str) -> dict:
    """
    Get the schema of a path.
    """

    if recap_struct := commands.schema(path):
        recap_dict = to_dict(recap_struct)
        if not isinstance(recap_dict, dict):
            raise HTTPException(
                status_code=503,
                detail=f"Expected a schema dict, but got {type(recap_dict)}",
            )
        return recap_dict
    raise HTTPException(status_code=404, detail="Path not found")



import re
from abc import ABC, abstractmethod
from typing import Any

from recap.types import RecapType, StructType


class DbapiConverter(ABC):
    def to_recap(self, columns: list[dict[str, Any]]) -> StructType:
        fields = []
        for column in columns:
            base_type = self._parse_type(column)
            is_nullable = column["IS_NULLABLE"].upper() == "YES"

            if is_nullable:
                base_type = base_type.make_nullable()

            if column["COLUMN_DEFAULT"] is not None or is_nullable:
                # TODO Convert default value to the correct type, not just a string
                base_type.extra_attrs["default"] = column["COLUMN_DEFAULT"]

            base_type.extra_attrs["name"] = column["COLUMN_NAME"]

            fields.append(base_type)

        return StructType(fields=fields)

    @abstractmethod
    def _parse_type(self, column_props: dict[str, Any]) -> RecapType:
        ...

    def _parse_parameters(self, col_type: str) -> list[str] | None:
        """
        Parse types that have parameters.
        """
        match = re.search(r"\((.*?)\)", col_type)
        if match:
            return [p.strip() for p in match.group(1).split(",")]
        return None

    def _get_time_unit(self, params: list[str] | None) -> str | None:
        match params:
            case [unit, *_] if int(unit) == 0:
                return "second"
            case [unit, *_] if int(unit) <= 3:
                return "millisecond"
            case [unit, *_] if int(unit) <= 6:
                return "microsecond"



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from recap.clients.dbapi import Connection, DbapiClient
from recap.converters.postgresql import PostgresqlConverter

PSYCOPG2_CONNECT_ARGS = {
    "host",
    "hostaddr",
    "port",
    "dbname",
    "user",
    "password",
    "passfile",
    "channel_binding",
    "connect_timeout",
    "client_encoding",
    "options",
    "application_name",
    "fallback_application_name",
    "keepalives",
    "keepalives_idle",
    "keepalives_interval",
    "keepalives_count",
    "tcp_user_timeout",
    "replication",
    "gssencmode",
    "sslmode",
    "requiressl",
    "sslcompression",
    "sslcert",
    "sslkey",
    "sslpassword",
    "sslrootcert",
    "sslcrl",
    "sslcrldir",
    "sslsni",
    "requirepeer",
    "ssl_min_protocol_version",
    "ssl_max_protocol_version",
    "krbsrvname",
    "gsslib",
    "service",
    "target_session_attrs",
}


class PostgresqlClient(DbapiClient):
    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, PostgresqlConverter())

    @staticmethod
    @contextmanager
    def create(
        paths: list[str] | None = None,
        **kwargs,
    ) -> Generator[PostgresqlClient, None, None]:
        import psycopg2

        if paths:
            kwargs["dbname"] = paths[0]

        # Only include kwargs that are valid for PsycoPG2 parse_dsn()
        kwargs = {k: v for k, v in kwargs.items() if k in PSYCOPG2_CONNECT_ARGS}

        with psycopg2.connect(**kwargs) as client:
            yield PostgresqlClient(client)

    def ls_catalogs(self) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute(
            """
            SELECT datname
            FROM pg_database
            ORDER BY datname ASC
            """
        )
        return [row[0] for row in cursor.fetchall()]



from confluent_kafka import schema_registry

from recap.clients import create_client
from recap.clients.confluent_registry import ConfluentRegistryClient
from recap.types import IntType, StringType, StructType, UnionType


class TestConfluentRegistryClient:
    @classmethod
    def setup_class(cls):
        cls.schema_registry_client = schema_registry.SchemaRegistryClient(
            {"url": "http://localhost:8081"}
        )

        # Define and register the Avro schema for the "dummy_topic"
        avro_schema_str = """
        {
            "type": "record",
            "name": "User",
            "fields": [
                {"name": "name", "type": "string"},
                {"name": "age",  "type": "int"}
            ]
        }
        """
        avro_schema = schema_registry.Schema(avro_schema_str, "AVRO")
        cls.schema_registry_client.register_schema("dummy_topic-value", avro_schema)

        # Define and register the Protobuf schema for the "dummy_topic_protobuf"
        protobuf_schema_str = """
        syntax = "proto3";

        message Person {
            string name = 1;
            int32 age = 2;
        }
        """
        protobuf_schema = schema_registry.Schema(protobuf_schema_str, "PROTOBUF")
        cls.schema_registry_client.register_schema(
            "dummy_topic_protobuf-value",
            protobuf_schema,
        )

        # Define and register the JSON schema for the "dummy_topic_json"
        json_schema_str = """
        {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": "User",
            "type": "object",
            "properties": {
                "name": {
                    "type": "string"
                },
                "age": {
                    "type": "integer"
                }
            },
            "required": ["name", "age"]
        }
        """
        json_schema = schema_registry.Schema(json_schema_str, "JSON")
        cls.schema_registry_client.register_schema(
            "dummy_topic_json-value",
            json_schema,
        )

    def test_struct_avro(self):
        client = ConfluentRegistryClient(self.schema_registry_client)
        result = client.get_schema("dummy_topic")

        assert isinstance(result, StructType)
        assert len(result.fields) == 2
        assert result.fields[0].extra_attrs["name"] == "name"
        assert isinstance(result.fields[0], StringType)
        assert result.fields[1].extra_attrs["name"] == "age"
        assert isinstance(result.fields[1], IntType)

    def test_struct_proto(self):
        client = ConfluentRegistryClient(self.schema_registry_client)
        result = client.get_schema("dummy_topic_protobuf")

        assert isinstance(result, StructType)
        assert len(result.fields) == 2
        assert result.fields[0].extra_attrs["name"] == "name"
        assert isinstance(result.fields[0], UnionType)
        assert isinstance(result.fields[0].types[1], StringType)
        assert result.fields[1].extra_attrs["name"] == "age"
        assert isinstance(result.fields[1], UnionType)
        assert isinstance(result.fields[1].types[1], IntType)

    def test_struct_json(self):
        client = ConfluentRegistryClient(self.schema_registry_client)
        result = client.get_schema("dummy_topic_json")

        assert isinstance(result, StructType)
        assert len(result.fields) == 2
        assert result.fields[0].extra_attrs["name"] == "name"
        assert isinstance(result.fields[0], StringType)
        assert result.fields[1].extra_attrs["name"] == "age"
        assert isinstance(result.fields[1], IntType)

    def test_create_client(self):
        csr_url = "http+csr://localhost:8081"

        with create_client(csr_url) as client:
            assert client.ls() == [
                "dummy_topic-value",
                "dummy_topic_json-value",
                "dummy_topic_protobuf-value",
            ]



import pytest
from pymetastore.htypes import (
    HCharType,
    HDecimalType,
    HListType,
    HMapType,
    HPrimitiveType,
    HStructType,
    HUnionType,
    HVarcharType,
    PrimitiveCategory,
)
from pymetastore.metastore import HColumn, HStorage, HTable, StorageFormat

from recap.converters.hive_metastore import HiveMetastoreConverter
from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


@pytest.mark.parametrize(
    "hive_type,expected",
    [
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.BOOLEAN),
            UnionType([NullType(), BoolType()], default=None, name="test_col"),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.BYTE),
            UnionType([NullType(), IntType(bits=8)], default=None, name="test_col"),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.SHORT),
            UnionType([NullType(), IntType(bits=16)], default=None, name="test_col"),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.INT),
            UnionType([NullType(), IntType(bits=32)], default=None, name="test_col"),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.LONG),
            UnionType([NullType(), IntType(bits=64)], default=None, name="test_col"),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.FLOAT),
            UnionType([NullType(), FloatType(bits=32)], default=None, name="test_col"),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.DOUBLE),
            UnionType([NullType(), FloatType(bits=64)], default=None, name="test_col"),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.STRING),
            UnionType(
                [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
                default=None,
                name="test_col",
            ),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.BINARY),
            UnionType(
                [NullType(), BytesType(bytes_=2_147_483_647)],
                default=None,
                name="test_col",
            ),
        ),
        (
            HDecimalType(precision=10, scale=2),
            UnionType(
                [
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=10,
                        scale=2,
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HVarcharType(length=255),
            UnionType(
                [NullType(), StringType(bytes_=255)],
                default=None,
                name="test_col",
            ),
        ),
        (
            HCharType(length=255),
            UnionType(
                [NullType(), StringType(bytes_=255, variable=False)],
                default=None,
                name="test_col",
            ),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.DATE),
            UnionType(
                [
                    NullType(),
                    IntType(
                        logical="build.recap.Date",
                        bits=32,
                        signed=True,
                        unit="day",
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.TIMESTAMP),
            UnionType(
                [
                    NullType(),
                    IntType(
                        logical="build.recap.Timestamp",
                        bits=64,
                        signed=True,
                        unit="nanosecond",
                        timezone="UTC",
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.TIMESTAMPLOCALTZ),
            UnionType(
                [
                    NullType(),
                    IntType(
                        logical="build.recap.Timestamp",
                        bits=64,
                        signed=True,
                        unit="nanosecond",
                        timezone=None,
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.INTERVAL_YEAR_MONTH),
            UnionType(
                [
                    NullType(),
                    BytesType(
                        logical="build.recap.Interval",
                        bytes_=12,
                        signed=True,
                        unit="month",
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HPrimitiveType(primitive_type=PrimitiveCategory.INTERVAL_DAY_TIME),
            UnionType(
                [
                    NullType(),
                    BytesType(
                        logical="build.recap.Interval",
                        bytes_=12,
                        signed=True,
                        unit="second",
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HMapType(
                key_type=HPrimitiveType(primitive_type=PrimitiveCategory.STRING),
                value_type=HPrimitiveType(primitive_type=PrimitiveCategory.INT),
            ),
            UnionType(
                [
                    NullType(),
                    MapType(
                        keys=UnionType(
                            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
                            default=None,
                        ),
                        values=UnionType(
                            [NullType(), IntType(bits=32)],
                            default=None,
                        ),
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HListType(
                element_type=HPrimitiveType(primitive_type=PrimitiveCategory.INT)
            ),
            UnionType(
                [
                    NullType(),
                    ListType(
                        values=UnionType(
                            [NullType(), IntType(bits=32)],
                            default=None,
                        ),
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HUnionType(
                types=[
                    HPrimitiveType(primitive_type=PrimitiveCategory.INT),
                    HPrimitiveType(primitive_type=PrimitiveCategory.STRING),
                ]
            ),
            UnionType(
                [
                    NullType(),
                    IntType(bits=32),
                    StringType(bytes_=9_223_372_036_854_775_807),
                ],
                default=None,
                name="test_col",
            ),
        ),
        (
            HStructType(
                names=["field1", "field2"],
                types=[
                    HPrimitiveType(primitive_type=PrimitiveCategory.INT),
                    HPrimitiveType(primitive_type=PrimitiveCategory.STRING),
                ],
            ),
            UnionType(
                [
                    NullType(),
                    StructType(
                        fields=[
                            UnionType(
                                [NullType(), IntType(bits=32)],
                                default=None,
                                name="field1",
                            ),
                            UnionType(
                                [
                                    NullType(),
                                    StringType(bytes_=9_223_372_036_854_775_807),
                                ],
                                default=None,
                                name="field2",
                            ),
                        ]
                    ),
                ],
                default=None,
                name="test_col",
            ),
        ),
    ],
)
def test_hive_metastore_converter(hive_type, expected):
    table = HTable(
        database_name="test_db",
        name="test_table",
        table_type="EXTERNAL_TABLE",
        columns=[HColumn(name="test_col", type=hive_type)],
        partition_columns=[],
        storage=HStorage(StorageFormat("", "", "")),
        parameters={},
    )

    result = HiveMetastoreConverter().to_recap(table)

    assert result.fields[0] == expected



import pytest

from recap.converters.mysql import MysqlConverter
from recap.types import BytesType, FloatType, IntType, StringType


@pytest.mark.parametrize(
    "column_props,expected",
    [
        (
            {
                "DATA_TYPE": "bigint",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=64, signed=True),
        ),
        (
            {
                "DATA_TYPE": "int",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=32, signed=True),
        ),
        (
            {
                "DATA_TYPE": "integer",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=32, signed=True),
        ),
        (
            {
                "DATA_TYPE": "mediumint",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=24, signed=True),
        ),
        (
            {
                "DATA_TYPE": "smallint",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=16, signed=True),
        ),
        (
            {
                "DATA_TYPE": "tinyint",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=8, signed=True),
        ),
        (
            {
                "DATA_TYPE": "double",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            FloatType(bits=64),
        ),
        (
            {
                "DATA_TYPE": "float",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 24,
                "NUMERIC_SCALE": None,
            },
            FloatType(bits=64),
        ),
        (
            {
                "DATA_TYPE": "float",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 23,
                "NUMERIC_SCALE": None,
            },
            FloatType(bits=32),
        ),
        (
            {
                "DATA_TYPE": "text",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "json",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "mediumtext",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "longtext",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "tinytext",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "varchar(100)",
                "CHARACTER_OCTET_LENGTH": 100,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=100, variable=True),
        ),
        (
            {
                "DATA_TYPE": "char(100)",
                "CHARACTER_OCTET_LENGTH": 100,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=100, variable=False),
        ),
        (
            {
                "DATA_TYPE": "enum",
                "CHARACTER_OCTET_LENGTH": 100,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=100, variable=False),
        ),
        (
            {
                "DATA_TYPE": "set",
                "CHARACTER_OCTET_LENGTH": 100,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=100, variable=False),
        ),
        (
            {
                "DATA_TYPE": "blob",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "mediumblob",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "longblob",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "tinyblob",
                "CHARACTER_OCTET_LENGTH": 1000,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=1000, variable=True),
        ),
        (
            {
                "DATA_TYPE": "varbinary(100)",
                "CHARACTER_OCTET_LENGTH": 100,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=100, variable=True),
        ),
        (
            {
                "DATA_TYPE": "binary(100)",
                "CHARACTER_OCTET_LENGTH": 100,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=100, variable=False),
        ),
        (
            {
                "DATA_TYPE": "bit(8",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
            },
            BytesType(bytes_=8, variable=False),
        ),
        (
            {
                "DATA_TYPE": "timestamp",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
                "DATETIME_PRECISION": 0,
            },
            IntType(bits=64, logical="build.recap.Timestamp", unit="second"),
        ),
        (
            {
                "DATA_TYPE": "datetime",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
                "DATETIME_PRECISION": 0,
            },
            IntType(bits=64, logical="build.recap.Timestamp", unit="second"),
        ),
        (
            {
                "DATA_TYPE": "dec",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=10,
                scale=2,
            ),
        ),
        (
            {
                "DATA_TYPE": "decimal",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=10,
                scale=2,
            ),
        ),
        (
            {
                "DATA_TYPE": "numeric",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=10,
                scale=2,
            ),
        ),
        (
            {
                "DATA_TYPE": "year",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=16, signed=False),
        ),
    ],
)
def test_mysql_converter(column_props, expected):
    result = MysqlConverter()._parse_type(column_props)
    assert result == expected



from typing import Annotated

import typer
import uvicorn
from rich import print_json

from recap import commands
from recap.settings import set_config, unset_config
from recap.types import to_dict

app = typer.Typer()


@app.command()
def ls(path: Annotated[str, typer.Argument(help="Path to list children of.")] = "/"):
    """
    List the children of a path.
    """

    if children := commands.ls(path):
        print_json(data=children)


@app.command()
def schema(path: Annotated[str, typer.Argument(help="Path to get schema of.")]):
    """
    Get the schema of a path.
    """

    if recap_struct := commands.schema(path):
        print_json(data=to_dict(recap_struct))


@app.command()
def add(
    system: Annotated[str, typer.Argument(help="User-defined name of the system.")],
    url: Annotated[str, typer.Argument(help="URL for the system.")],
):
    """
    Add a system to the config.
    """

    set_config(f"RECAP_SYSTEMS__{system}", url)


@app.command()
def remove(
    system: Annotated[str, typer.Argument(help="User-defined name of the system.")],
):
    """
    Remove a system from the config.
    """

    unset_config(f"RECAP_SYSTEMS__{system}")


@app.command()
def serve(
    host: str = "127.0.0.1",
    port: int = 8000,
    log_level: str = "info",
):
    """
    Serve the API.
    """

    uvicorn.run("recap.gateway:app", host=host, port=port, log_level=log_level)



from json import loads

import pytest
from jsonschema.validators import Draft202012Validator

from recap.converters.json_schema import JSONSchemaConverter
from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    StringType,
    StructType,
    UnionType,
)


def test_all_basic_types():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {"type": "string"},
            "a_number": {"type": "number"},
            "an_integer": {"type": "integer"},
            "a_boolean": {"type": "boolean"},
            "a_null": {"type": "null"}
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema))
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
            name="a_string",
            default=None,
        ),
        UnionType(
            [NullType(), FloatType(bits=64)],
            name="a_number",
            default=None,
        ),
        UnionType(
            [NullType(), IntType(bits=32, signed=True)],
            name="an_integer",
            default=None,
        ),
        UnionType(
            [NullType(), BoolType()],
            name="a_boolean",
            default=None,
        ),
        UnionType(
            [NullType(), NullType()],
            name="a_null",
            default=None,
        ),
    ]


def test_nested_objects():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "an_object": {
                "type": "object",
                "properties": {
                    "a_string": {"type": "string"}
                }
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema))
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [
                NullType(),
                StructType(
                    [
                        UnionType(
                            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
                            name="a_string",
                            default=None,
                        ),
                    ],
                ),
            ],
            name="an_object",
            default=None,
        ),
    ]


def test_object_with_array_of_objects():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "an_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "a_string": {"type": "string"}
                    }
                }
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema))
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [
                NullType(),
                ListType(
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    StringType(bytes_=9_223_372_036_854_775_807),
                                ],
                                name="a_string",
                                default=None,
                            )
                        ]
                    ),
                ),
            ],
            name="an_array",
            default=None,
        ),
    ]


def test_required_properties():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "required_property": {"type": "string"},
            "optional_property": {"type": "number"}
        },
        "required": ["required_property"]
    }
    """
    Draft202012Validator.check_schema(loads(json_schema))
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        StringType(bytes_=9_223_372_036_854_775_807, name="required_property"),
        UnionType(
            [NullType(), FloatType(bits=64)],
            name="optional_property",
            default=None,
        ),
    ]


def test_doc_attribute():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {
                "type": "string",
                "description": "This is a string"
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema))
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
            name="a_string",
            doc="This is a string",
            default=None,
        ),
    ]


def test_name_attribute():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {
                "type": "string",
                "title": "A String"
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema))
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
            name="a_string",
            default=None,
        ),
    ]


def test_default_attribute():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {
                "type": "string",
                "default": "Default value"
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema))
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [
                NullType(),
                StringType(bytes_=9_223_372_036_854_775_807),
            ],
            name="a_string",
            default="Default value",
        ),
    ]


def test_convert_bytes():
    schema = """
    {
        "type": "object",
        "properties": {
            "img": {
                "type": "string",
                "format": "bytes"
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(schema))
    result = JSONSchemaConverter().to_recap(schema)
    assert result == StructType(
        [
            UnionType(
                [NullType(), BytesType(bytes_=9_223_372_036_854_775_807)],
                name="img",
                default=None,
            ),
        ]
    )


def test_convert_date():
    converter = JSONSchemaConverter()
    schema = """
    {
        "type": "object",
        "properties": {
            "date": {
                "type": "string",
                "format": "date"
            }
        },
        "required": ["date"]
    }
    """
    Draft202012Validator.check_schema(loads(schema))
    result = converter.to_recap(schema)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], StringType)
    assert result.fields[0].logical == "org.iso.8601.Date"


def test_convert_datetime():
    converter = JSONSchemaConverter()
    schema = """
    {
        "type": "object",
        "properties": {
            "datetime": {
                "type": "string",
                "format": "date-time"
            }
        },
        "required": ["datetime"]
    }
    """
    Draft202012Validator.check_schema(loads(schema))
    result = converter.to_recap(schema)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], StringType)
    assert result.fields[0].logical == "org.iso.8601.DateTime"


def test_convert_time():
    converter = JSONSchemaConverter()
    schema = """
    {
        "type": "object",
        "properties": {
            "time": {
                "type": "string",
                "format": "time"
            }
        },
        "required": ["time"]
    }
    """
    Draft202012Validator.check_schema(loads(schema))
    result = converter.to_recap(schema)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], StringType)
    assert result.fields[0].logical == "org.iso.8601.Time"


def test_id_to_recap_alias():
    json_schema_str = """
    {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "http://example.com/path/to/schema",
        "type": "object",
        "properties": {
            "field": {
                "type": "string"
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema_str))
    recap_type = JSONSchemaConverter().to_recap(json_schema_str)
    assert recap_type == StructType(
        fields=[
            UnionType(
                [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
                name="field",
                default=None,
            )
        ],
        alias="com.example.path.to.schema",
    )


def test_id_to_recap_alias_schema_default():
    json_schema_str = """
    {
        "$id": "http://example.com/path/to/schema",
        "type": "object",
        "properties": {
            "field": {
                "type": "string"
            }
        }
    }
    """
    Draft202012Validator.check_schema(loads(json_schema_str))
    recap_type = JSONSchemaConverter().to_recap(json_schema_str)
    assert recap_type == StructType(
        fields=[
            UnionType(
                [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
                name="field",
                default=None,
            )
        ],
        alias="com.example.path.to.schema",
    )


def test_cyclic_reference():
    json_schema_str = """
    {
        "$id": "http://recap.build/linkedlist.schema.json",
        "description": "A node in a linked list",
        "type": "object",
        "properties": {
            "value": {
                "description": "Value of the node",
                "type": "integer"
            },
            "next": {
                "description": "Next node in the linked list",
                "$ref": "http://recap.build/linkedlist.schema.json"
            }
        },
        "required": ["value"]
    }
    """
    Draft202012Validator.check_schema(loads(json_schema_str))
    converter = JSONSchemaConverter()
    recap_type = converter.to_recap(json_schema_str)
    assert recap_type == StructType(
        fields=[
            IntType(
                name="value",
                doc="Value of the node",
                bits=32,
            ),
            UnionType(
                [
                    NullType(),
                    ProxyType(
                        alias="build.recap.linkedlist.schema.json",
                        registry=converter.registry,
                    ),
                ],
                doc="Next node in the linked list",
                name="next",
                default=None,
            ),
        ],
        alias="build.recap.linkedlist.schema.json",
        doc="A node in a linked list",
    )


def test_cyclic_reference_dynamic():
    json_schema_str = """
    {
        "$id": "http://recap.build/linkedlist.schema.json",
        "description": "A node in a linked list",
        "type": "object",
        "properties": {
            "value": {
                "description": "Value of the node",
                "type": "integer"
            },
            "next": {
                "description": "Next node in the linked list",
                "$ref": "http://recap.build/linkedlist.schema.json"
            }
        },
        "required": ["value"]
    }
    """
    Draft202012Validator.check_schema(loads(json_schema_str))
    converter = JSONSchemaConverter()
    recap_type = converter.to_recap(json_schema_str)
    assert recap_type == StructType(
        fields=[
            IntType(
                name="value",
                doc="Value of the node",
                bits=32,
            ),
            UnionType(
                [
                    NullType(),
                    ProxyType(
                        alias="build.recap.linkedlist.schema.json",
                        registry=converter.registry,
                    ),
                ],
                doc="Next node in the linked list",
                name="next",
                default=None,
            ),
        ],
        alias="build.recap.linkedlist.schema.json",
        doc="A node in a linked list",
    )


@pytest.mark.xfail(reason="Defs aren't currently implemented")
def test_defs_reference():
    json_schema_str = """
    {
        "type": "object",
        "$id": "http://recap.build/linkedlist.schema.json",
        "properties": {
            "defRef": { "$ref": "#/$defs/positiveInteger" }
        },
        "$defs": {
            "positiveInteger": {
                "type": "integer"
            }
        },
        "required": ["defRef"]
    }
    """
    Draft202012Validator.check_schema(loads(json_schema_str))
    converter = JSONSchemaConverter()
    recap_type = converter.to_recap(json_schema_str)
    assert recap_type == StructType(
        fields=[
            ProxyType(
                alias="build.recap.linkedlist.schema.json",
                registry=converter.registry,
                name="defRef",
            ),
        ],
        alias="build.recap.linkedlist.schema.json",
    )


def test_from_recap_types():
    struct = StructType(
        [
            IntType(bits=32, name="int_field"),
            StringType(name="string_field"),
            BytesType(name="bytes_field"),
            FloatType(bits=64, name="float_field"),
            BoolType(name="bool_field"),
            NullType(name="null_field"),
            ListType(values=IntType(bits=32), name="list_field"),
            MapType(keys=StringType(), values=IntType(bits=32), name="map_field"),
            EnumType(symbols=["A", "B", "C"], name="enum_field"),
            StructType(
                fields=[IntType(bits=32, name="nested_int_field")],
                name="nested_struct_field",
            ),
        ]
    )

    schema = JSONSchemaConverter().from_recap(struct)
    expected_schema = {
        "type": "object",
        "properties": {
            "int_field": {"type": "integer"},
            "string_field": {"type": "string"},
            "bytes_field": {"type": "string", "format": "byte"},
            "float_field": {"type": "number"},
            "bool_field": {"type": "boolean"},
            "null_field": {"type": "null"},
            "list_field": {"type": "array", "items": {"type": "integer"}},
            "map_field": {
                "type": "object",
                "additionalProperties": {"type": "integer"},
            },
            "enum_field": {"type": "string", "enum": ["A", "B", "C"]},
            "nested_struct_field": {
                "type": "object",
                "properties": {
                    "nested_int_field": {"type": "integer"},
                },
                "required": ["nested_int_field"],
            },
        },
        "required": [
            "int_field",
            "string_field",
            "bytes_field",
            "float_field",
            "bool_field",
            "null_field",
            "list_field",
            "map_field",
            "enum_field",
            "nested_struct_field",
        ],
    }

    Draft202012Validator.check_schema(expected_schema)
    assert schema == expected_schema


def test_from_recap_optional_types():
    struct = StructType(
        [
            UnionType(
                [NullType(), IntType(bits=32)],
                name="optional_int_field",
                default=0,
            ),
            UnionType(
                [NullType(), StringType()],
                name="optional_string_field",
                default="",
            ),
            UnionType(
                [NullType(), BytesType()],
                name="optional_bytes_field",
                default=b"",
            ),
            UnionType(
                [NullType(), FloatType(bits=64)],
                name="optional_float_field",
                default=0.0,
            ),
            UnionType(
                [NullType(), BoolType()],
                name="optional_bool_field",
                default=False,
            ),
            UnionType(
                [NullType(), ListType(values=IntType(bits=32))],
                name="optional_list_field",
                default=[],
            ),
            UnionType(
                [NullType(), MapType(keys=StringType(), values=IntType(bits=32))],
                name="optional_map_field",
                default={},
            ),
            UnionType(
                [NullType(), EnumType(symbols=["A", "B", "C"])],
                name="optional_enum_field",
                default="A",
            ),
            UnionType(
                [
                    NullType(),
                    StructType(
                        fields=[
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32),
                                ],
                                default=None,
                                name="nested_optional_int_field",
                            )
                        ],
                        name="struct",
                    ),
                ],
                name="optional_nested_struct_field",
                default={"nested_optional_int_field": 0},
            ),
        ]
    )

    schema = JSONSchemaConverter().from_recap(struct)
    expected_schema = {
        "type": "object",
        "properties": {
            "optional_int_field": {"type": "integer", "default": 0},
            "optional_string_field": {"type": "string", "default": ""},
            "optional_bytes_field": {
                "type": "string",
                "format": "byte",
                "default": b"",
            },
            "optional_float_field": {"type": "number", "default": 0.0},
            "optional_bool_field": {"type": "boolean", "default": False},
            "optional_list_field": {
                "type": "array",
                "items": {"type": "integer"},
                "default": [],
            },
            "optional_map_field": {
                "type": "object",
                "additionalProperties": {"type": "integer"},
                "default": {},
            },
            "optional_enum_field": {
                "type": "string",
                "enum": ["A", "B", "C"],
                "default": "A",
            },
            "optional_nested_struct_field": {
                "type": "object",
                "properties": {
                    "nested_optional_int_field": {"type": "integer", "default": None},
                },
                "default": {"nested_optional_int_field": 0},
            },
        },
    }

    Draft202012Validator.check_schema(expected_schema)
    assert schema == expected_schema


def test_from_recap_types_with_aliases():
    struct = StructType(
        [
            IntType(bits=32, name="int_field", alias="build.recap.MyInt"),
            StringType(name="string_field", alias="build.recap.MyString"),
            BytesType(name="bytes_field", alias="build.recap.MyBytes"),
            FloatType(bits=64, name="float_field", alias="build.recap.MyFloat"),
            BoolType(name="bool_field", alias="build.recap.MyBool"),
            NullType(name="null_field", alias="build.recap.MyNull"),
            ListType(
                values=IntType(bits=32, alias="build.recap.ListItem"),
                name="list_field",
                alias="build.recap.MyList",
            ),
            MapType(
                keys=StringType(alias="build.recap.MapKey"),
                values=IntType(bits=32, alias="build.recap.MapValue"),
                name="map_field",
                alias="build.recap.MyMap",
            ),
            EnumType(
                symbols=["A", "B", "C"], name="enum_field", alias="build.recap.MyEnum"
            ),
            StructType(
                fields=[
                    IntType(
                        bits=32, name="nested_int_field", alias="build.recap.NestedInt"
                    )
                ],
                name="nested_struct_field",
                alias="build.recap.NestedStruct",
            ),
        ]
    )

    schema = JSONSchemaConverter().from_recap(struct)

    expected_schema = {
        "$defs": {
            "build.recap.MyInt": {"type": "integer"},
            "build.recap.MyString": {"type": "string"},
            "build.recap.MyBytes": {"type": "string", "format": "byte"},
            "build.recap.MyFloat": {"type": "number"},
            "build.recap.MyBool": {"type": "boolean"},
            "build.recap.MyNull": {"type": "null"},
            "build.recap.ListItem": {"type": "integer"},
            "build.recap.MyList": {
                "type": "array",
                "items": {"$ref": "#/$defs/build.recap.ListItem"},
            },
            "build.recap.MapValue": {"type": "integer"},
            "build.recap.MyMap": {
                "type": "object",
                "additionalProperties": {"$ref": "#/$defs/build.recap.MapValue"},
            },
            "build.recap.MyEnum": {"type": "string", "enum": ["A", "B", "C"]},
            "build.recap.NestedInt": {"type": "integer"},
            "build.recap.NestedStruct": {
                "type": "object",
                "properties": {
                    "nested_int_field": {"$ref": "#/$defs/build.recap.NestedInt"},
                },
                "required": ["nested_int_field"],
            },
        },
        "type": "object",
        "properties": {
            "int_field": {"$ref": "#/$defs/build.recap.MyInt"},
            "string_field": {"$ref": "#/$defs/build.recap.MyString"},
            "bytes_field": {"$ref": "#/$defs/build.recap.MyBytes"},
            "float_field": {"$ref": "#/$defs/build.recap.MyFloat"},
            "bool_field": {"$ref": "#/$defs/build.recap.MyBool"},
            "null_field": {"$ref": "#/$defs/build.recap.MyNull"},
            "list_field": {"$ref": "#/$defs/build.recap.MyList"},
            "map_field": {"$ref": "#/$defs/build.recap.MyMap"},
            "enum_field": {"$ref": "#/$defs/build.recap.MyEnum"},
            "nested_struct_field": {"$ref": "#/$defs/build.recap.NestedStruct"},
        },
        "required": [
            "int_field",
            "string_field",
            "bytes_field",
            "float_field",
            "bool_field",
            "null_field",
            "list_field",
            "map_field",
            "enum_field",
            "nested_struct_field",
        ],
    }

    Draft202012Validator.check_schema(expected_schema)
    assert schema == expected_schema


def test_from_recap_types_alias_with_optional_field():
    struct = ListType(
        values=StructType(
            fields=[
                UnionType(
                    alias="build.recap.enabledToggle",
                    types=[NullType(), BoolType()],
                    name="enabled",
                    default=None,
                ),
            ],
        ),
    )

    schema = JSONSchemaConverter().from_recap(struct)

    expected_schema = {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "enabled": {
                    "$ref": "#/$defs/build.recap.enabledToggle",
                }
            },
        },
        "$defs": {
            "build.recap.enabledToggle": {
                "type": "boolean",
                "default": None,
            }
        },
    }

    Draft202012Validator.check_schema(expected_schema)
    assert schema == expected_schema


def test_from_recap_types_alias_with_optional_non_null():
    struct = ListType(
        values=StructType(
            fields=[
                UnionType(
                    alias="build.recap.enabledToggle",
                    types=[NullType(), BoolType()],
                    name="enabled",
                    default=True,
                ),
            ],
        ),
    )

    schema = JSONSchemaConverter().from_recap(struct)

    expected_schema = {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "enabled": {
                    "$ref": "#/$defs/build.recap.enabledToggle",
                }
            },
        },
        "$defs": {
            "build.recap.enabledToggle": {
                "type": "boolean",
                "default": True,
            }
        },
    }

    Draft202012Validator.check_schema(expected_schema)
    assert schema == expected_schema


def test_from_recap_types_union_type():
    struct = StructType(
        fields=[
            UnionType(
                alias="build.recap.UnionExample",
                types=[IntType(bits=32), StringType()],
                name="example_field",
            ),
        ],
    )

    schema = JSONSchemaConverter().from_recap(struct)

    expected_schema = {
        "type": "object",
        "properties": {
            "example_field": {
                "$ref": "#/$defs/build.recap.UnionExample",
            },
        },
        "$defs": {
            "build.recap.UnionExample": {
                "oneOf": [
                    {
                        "type": "integer",
                    },
                    {
                        "type": "string",
                    },
                ],
            },
        },
        "required": ["example_field"],
    }

    Draft202012Validator.check_schema(expected_schema)
    assert schema == expected_schema



import pytest

from recap.converters.snowflake import SnowflakeConverter
from recap.types import BoolType, BytesType, FloatType, IntType, StringType


@pytest.mark.parametrize(
    "column_props,expected_type",
    [
        # float types
        ({"DATA_TYPE": "float", "CHARACTER_OCTET_LENGTH": None}, FloatType(bits=64)),
        ({"DATA_TYPE": "float4", "CHARACTER_OCTET_LENGTH": None}, FloatType(bits=64)),
        ({"DATA_TYPE": "float8", "CHARACTER_OCTET_LENGTH": None}, FloatType(bits=64)),
        ({"DATA_TYPE": "double", "CHARACTER_OCTET_LENGTH": None}, FloatType(bits=64)),
        (
            {"DATA_TYPE": "double precision", "CHARACTER_OCTET_LENGTH": None},
            FloatType(bits=64),
        ),
        ({"DATA_TYPE": "real", "CHARACTER_OCTET_LENGTH": None}, FloatType(bits=64)),
        ({"DATA_TYPE": "boolean", "CHARACTER_OCTET_LENGTH": None}, BoolType()),
        (
            {
                "DATA_TYPE": "number",
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
                "CHARACTER_OCTET_LENGTH": None,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=16,
                variable=False,
                precision=10,
                scale=2,
            ),
        ),
        (
            {
                "DATA_TYPE": "decimal",
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
                "CHARACTER_OCTET_LENGTH": None,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=16,
                variable=False,
                precision=10,
                scale=2,
            ),
        ),
        (
            {
                "DATA_TYPE": "numeric",
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
                "CHARACTER_OCTET_LENGTH": None,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=16,
                variable=False,
                precision=10,
                scale=2,
            ),
        ),
        (
            {"DATA_TYPE": "varchar", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        (
            {"DATA_TYPE": "nvarchar", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        (
            {"DATA_TYPE": "nvarchar2", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        (
            {"DATA_TYPE": "char varying", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        (
            {"DATA_TYPE": "nchar varying", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        (
            {"DATA_TYPE": "char", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        (
            {"DATA_TYPE": "nchar", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        (
            {"DATA_TYPE": "character", "CHARACTER_OCTET_LENGTH": 5},
            StringType(bytes_=5, variable=True),
        ),
        ({"DATA_TYPE": "binary", "CHARACTER_OCTET_LENGTH": 5}, BytesType(bytes_=5)),
        ({"DATA_TYPE": "varbinary", "CHARACTER_OCTET_LENGTH": 5}, BytesType(bytes_=5)),
        ({"DATA_TYPE": "blob", "CHARACTER_OCTET_LENGTH": 5}, BytesType(bytes_=5)),
        (
            {"DATA_TYPE": "date", "CHARACTER_OCTET_LENGTH": None},
            IntType(bits=32, logical="build.recap.Date", unit="day"),
        ),
        (
            {"DATA_TYPE": "timestamp(3)", "CHARACTER_OCTET_LENGTH": None},
            IntType(bits=64, logical="build.recap.Timestamp", unit="millisecond"),
        ),
        (
            {"DATA_TYPE": "timestamp", "CHARACTER_OCTET_LENGTH": None},
            IntType(bits=64, logical="build.recap.Timestamp", unit="nanosecond"),
        ),
        (
            {"DATA_TYPE": "time", "CHARACTER_OCTET_LENGTH": None},
            IntType(bits=32, logical="build.recap.Time", unit="nanosecond"),
        ),
    ],
)
def test_snowflake_converter(column_props, expected_type):
    assert SnowflakeConverter()._parse_type(column_props) == expected_type



import pytest
from google.api_core.client_options import ClientOptions
from google.auth.credentials import AnonymousCredentials
from google.cloud import bigquery
from google.cloud.bigquery import QueryJobConfig

from recap.clients.bigquery import BigQueryClient
from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


@pytest.fixture(scope="module")
def client():
    client_options = ClientOptions(api_endpoint="http://localhost:9050")
    client = bigquery.Client(
        "test_project",
        client_options=client_options,
        credentials=AnonymousCredentials(),
    )
    return client


@pytest.fixture(scope="module", autouse=True)
def setup_data(client):
    # Creating a table with all primitive types.
    client.query(
        query="""
        CREATE TABLE `test_dataset.test_table` (
            test_string STRING,
            test_bytes BYTES,
            test_int64 INT64,
            test_float64 FLOAT64,
            test_boolean BOOLEAN,
            test_timestamp TIMESTAMP,
            test_datetime DATETIME,
            test_date DATE,
            test_time TIME,
            test_numeric NUMERIC,
            test_bigdecimal BIGNUMERIC
        )
        """,
        job_config=QueryJobConfig(),
    )

    # Creating a table with REQUIRED primitive types.
    client.query(
        query="""
        CREATE TABLE `test_dataset.test_table_required` (
            test_string STRING NOT NULL,
            test_bytes BYTES NOT NULL,
            test_int64 INT64 NOT NULL,
            test_float64 FLOAT64 NOT NULL,
            test_boolean BOOLEAN NOT NULL,
            test_timestamp TIMESTAMP NOT NULL,
            test_datetime DATETIME NOT NULL,
            test_date DATE NOT NULL,
            test_time TIME NOT NULL,
            test_numeric NUMERIC NOT NULL,
            test_bigdecimal BIGNUMERIC NOT NULL
        )
        """,
        job_config=QueryJobConfig(),
    )

    # Creating a table with a nested STRUCT / RECORD type.
    client.query(
        query="""
        CREATE TABLE `test_dataset.test_table_struct` (
            test_record STRUCT<
                test_boolean BOOLEAN,
                nested_record STRUCT<
                    test_nested_boolean BOOLEAN
                >
            >
        )
        """,
        job_config=QueryJobConfig(),
    )

    # Creating a table with a repeated (array) type.
    client.query(
        query="""
        CREATE TABLE `test_dataset.test_table_repeated` (
            test_array ARRAY<STRING>
        )
        """,
        job_config=QueryJobConfig(),
    )

    # Creating a table with a repeated RECORD type.
    client.query(
        query="""
        CREATE TABLE `test_dataset.test_table_repeated_records` (
            test_array ARRAY<STRUCT<test_boolean BOOLEAN, test_int64 INT64>>
        )
        """,
        job_config=QueryJobConfig(),
    )

    # Creating a table with a default value.
    client.query(
        query="""
        CREATE TABLE `test_dataset.test_table_default` (
            test_string STRING DEFAULT "default_value"
        )
        """,
        job_config=QueryJobConfig(),
    )

    # Creating a table with a column description.
    client.query(
        query="""
        CREATE TABLE `test_dataset.test_table_description` (
            test_string STRING OPTIONS(description="This is a test string column")
        )
        """,
        job_config=QueryJobConfig(),
    )


def test_primitive_types(client):
    client = BigQueryClient(client)
    recap_schema = client.get_schema("test_project", "test_dataset", "test_table")
    recap_fields = recap_schema.fields

    assert recap_fields[0] == UnionType(
        types=[NullType(), StringType()],
        default=None,
        name="test_string",
    )

    assert recap_fields[1] == UnionType(
        types=[NullType(), BytesType()],
        default=None,
        name="test_bytes",
    )

    assert recap_fields[2] == UnionType(
        types=[NullType(), IntType(bits=64)],
        default=None,
        name="test_int64",
    )

    assert recap_fields[3] == UnionType(
        types=[NullType(), FloatType(bits=64)],
        default=None,
        name="test_float64",
    )

    assert recap_fields[4] == UnionType(
        types=[NullType(), BoolType()],
        default=None,
        name="test_boolean",
    )

    assert recap_fields[5] == UnionType(
        types=[
            NullType(),
            IntType(
                bits=64,
                logical="build.recap.Timestamp",
                unit="microsecond",
            ),
        ],
        default=None,
        name="test_timestamp",
    )

    assert recap_fields[6] == UnionType(
        types=[
            NullType(),
            IntType(
                bits=64,
                logical="build.recap.Timestamp",
                unit="microsecond",
            ),
        ],
        default=None,
        name="test_datetime",
    )

    assert recap_fields[7] == UnionType(
        types=[
            NullType(),
            IntType(
                bits=32,
                logical="build.recap.Date",
                unit="day",
            ),
        ],
        default=None,
        name="test_date",
    )

    assert recap_fields[8] == UnionType(
        types=[
            NullType(),
            IntType(
                bits=32,
                logical="build.recap.Time",
                unit="microsecond",
            ),
        ],
        default=None,
        name="test_time",
    )

    assert recap_fields[9] == UnionType(
        types=[
            NullType(),
            BytesType(
                bytes_=16,
                variable=False,
                logical="build.recap.Decimal",
                precision=38,
                scale=0,
            ),
        ],
        default=None,
        name="test_numeric",
    )

    assert recap_fields[10] == UnionType(
        types=[
            NullType(),
            BytesType(
                bytes_=32,
                variable=False,
                logical="build.recap.Decimal",
                precision=76,
                scale=0,
            ),
        ],
        default=None,
        name="test_bigdecimal",
    )


# TODO Remove xfail after https://github.com/goccy/bigquery-emulator/issues/210
@pytest.mark.xfail(reason="BigQuery emulator does not support REQUIRED fields")
def test_required_types(client):
    client = BigQueryClient(client)
    recap_schema = client.get_schema(
        "test_project",
        "test_dataset",
        "test_table_required",
    )
    recap_fields = recap_schema.fields

    assert recap_fields[0] == StringType(name="test_string")

    assert recap_fields[1] == BytesType(name="test_bytes")

    assert recap_fields[2] == IntType(bits=64, name="test_int64")

    assert recap_fields[3] == FloatType(bits=64, name="test_float64")

    assert recap_fields[4] == BoolType(name="test_boolean")

    assert recap_fields[5] == IntType(
        bits=64,
        logical="build.recap.Timestamp",
        unit="microsecond",
        name="test_timestamp",
    )

    assert recap_fields[6] == IntType(
        bits=64,
        logical="build.recap.Timestamp",
        unit="microsecond",
        name="test_datetime",
    )

    assert recap_fields[7] == IntType(
        bits=32,
        logical="build.recap.Date",
        unit="day",
        name="test_date",
    )

    assert recap_fields[8] == IntType(
        bits=32,
        logical="build.recap.Time",
        unit="microsecond",
        name="test_time",
    )

    assert recap_fields[9] == BytesType(
        bytes_=16,
        variable=False,
        logical="build.recap.Decimal",
        precision=38,
        scale=0,
        name="test_numeric",
    )

    assert recap_fields[10] == BytesType(
        bytes_=32,
        variable=False,
        logical="build.recap.Decimal",
        precision=76,
        scale=0,
        name="test_bigdecimal",
    )


def test_nested_struct_record_types(client):
    client = BigQueryClient(client)
    recap_schema = client.get_schema(
        "test_project",
        "test_dataset",
        "test_table_struct",
    )
    recap_fields = recap_schema.fields

    assert recap_fields[0] == UnionType(
        types=[
            NullType(),
            StructType(
                fields=[
                    UnionType(
                        types=[
                            NullType(),
                            BoolType(),
                        ],
                        default=None,
                        name="test_boolean",
                    ),
                    UnionType(
                        types=[
                            NullType(),
                            StructType(
                                fields=[
                                    UnionType(
                                        types=[
                                            NullType(),
                                            BoolType(),
                                        ],
                                        default=None,
                                        name="test_nested_boolean",
                                    ),
                                ],
                            ),
                        ],
                        default=None,
                        name="nested_record",
                    ),
                ],
            ),
        ],
        default=None,
        name="test_record",
    )


def test_repeated_types(client):
    client = BigQueryClient(client)
    recap_schema = client.get_schema(
        "test_project",
        "test_dataset",
        "test_table_repeated",
    )
    recap_fields = recap_schema.fields

    # ARRAYs can't be nullable and their elements can't be nullable. See:
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#array_nulls

    assert recap_fields[0] == ListType(
        values=StringType(),
        name="test_array",
    )


def test_repeated_records(client):
    client = BigQueryClient(client)
    recap_schema = client.get_schema(
        "test_project",
        "test_dataset",
        "test_table_repeated_records",
    )
    recap_fields = recap_schema.fields

    assert recap_fields[0] == ListType(
        values=StructType(
            fields=[
                UnionType(
                    types=[
                        NullType(),
                        BoolType(),
                    ],
                    default=None,
                    name="test_boolean",
                ),
                UnionType(
                    types=[
                        NullType(),
                        IntType(bits=64),
                    ],
                    default=None,
                    name="test_int64",
                ),
            ],
        ),
        name="test_array",
    )


# TODO Remove xfail after https://github.com/goccy/bigquery-emulator/issues/211
@pytest.mark.xfail(reason="BigQuery emulator does not support DEFAULT values")
def test_default_value(client):
    client = BigQueryClient(client)
    recap_schema = client.get_schema(
        "test_project",
        "test_dataset",
        "test_table_default",
    )
    recap_fields = recap_schema.fields

    assert recap_fields[0] == UnionType(
        types=[NullType(), StringType()],
        default="default_value",
        name="test_string",
    )


# TODO Remove xfail after https://github.com/goccy/bigquery-emulator/issues/212
@pytest.mark.xfail(reason="BigQuery emulator does not support OPTIONS")
def test_column_description(client):
    client = BigQueryClient(client)
    recap_schema = client.get_schema(
        "test_project",
        "test_dataset",
        "test_table_description",
    )
    recap_fields = recap_schema.fields

    assert recap_fields[0] == UnionType(
        types=[NullType(), StringType()],
        default=None,
        name="test_string",
        docs="This is a test string column",
    )


def test_ls(client):
    client = BigQueryClient(client)
    # TODO BigQuery emulator doesn't support ls_projects()
    # assert client.ls() == ["test_project"]
    assert client.ls("test_project") == ["test_dataset"]
    assert client.ls("test_project", "test_dataset") == [
        "test_table",
        "test_table_required",
        "test_table_struct",
        "test_table_repeated",
        "test_table_repeated_records",
        # TODO Enable after https://github.com/goccy/bigquery-emulator/issues/211
        # "test_table_default",
        "test_table_description",
    ]



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from recap.clients.dbapi import Connection, DbapiClient
from recap.converters.snowflake import SnowflakeConverter


class SnowflakeClient(DbapiClient):
    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, SnowflakeConverter())

    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[SnowflakeClient, None, None]:
        import snowflake.connector

        with snowflake.connector.connect(**kwargs) as client:
            yield SnowflakeClient(client)  # type: ignore



from __future__ import annotations

from pymetastore.htypes import (
    HCharType,
    HDecimalType,
    HListType,
    HMapType,
    HPrimitiveType,
    HStructType,
    HType,
    HUnionType,
    HVarcharType,
    PrimitiveCategory,
)
from pymetastore.metastore import HColumn, HTable

from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    RecapType,
    StringType,
    StructType,
    UnionType,
)


class HiveMetastoreConverter:
    def to_recap(
        self,
        table: HTable,
    ) -> StructType:
        fields: list[RecapType] = []
        for col in table.columns:
            recap_type = self._parse_schema(col)
            fields.append(recap_type)
        return StructType(fields=fields, name=table.name)

    def _parse_schema(self, htype: HType | HColumn) -> RecapType:
        recap_type: RecapType | None = None
        extra_attrs = {}
        if isinstance(htype, HColumn):
            extra_attrs["name"] = htype.name
            if htype.comment:
                extra_attrs["doc"] = htype.comment
            htype = htype.type
        match htype:
            case HPrimitiveType(primitive_type=PrimitiveCategory.BOOLEAN):
                recap_type = BoolType(**extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.BYTE):
                recap_type = IntType(bits=8, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.SHORT):
                recap_type = IntType(bits=16, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.INT):
                recap_type = IntType(bits=32, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.LONG):
                recap_type = IntType(bits=64, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.FLOAT):
                recap_type = FloatType(bits=32, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.DOUBLE):
                recap_type = FloatType(bits=64, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.VOID):
                recap_type = NullType(**extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.STRING):
                # TODO: Should handle multi-byte encodings
                # Using 2^63-1 as the max length because Hive has no defined max.
                recap_type = StringType(bytes_=9_223_372_036_854_775_807, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.BINARY):
                recap_type = BytesType(bytes_=2_147_483_647, **extra_attrs)
            case HDecimalType(precision=int(precision), scale=int(scale)):
                # TODO: Decimals can be > 38 digits, but need to calculate bytes_.
                if precision > 38:
                    raise ValueError(f"Max decimal precision is 38, got {precision}")
                recap_type = BytesType(
                    logical="build.recap.Decimal",
                    bytes_=16,
                    variable=False,
                    precision=precision,
                    scale=scale,
                    **extra_attrs,
                )
            case HVarcharType(length=int(length)):
                # TODO: Should handle multi-byte encodings
                recap_type = StringType(bytes_=length, **extra_attrs)
            case HCharType(length=int(length)):
                # TODO: Should handle multi-byte encodings
                recap_type = StringType(bytes_=length, variable=False, **extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.DATE):
                recap_type = IntType(
                    logical="build.recap.Date",
                    bits=32,
                    signed=True,
                    unit="day",
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.TIMESTAMP):
                # Hive stores timestamps as strings, so 64 bits isn't enough to store
                # all possible values. C'est la vie.
                recap_type = IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="nanosecond",
                    timezone="UTC",
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.TIMESTAMPLOCALTZ):
                # Hive stores timestamps as strings, so 64 bits isn't enough to store
                # all possible values. C'est la vie.
                recap_type = IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="nanosecond",
                    # Follow Apache Arrow's Schema.fbs style for local time.
                    # Denote LOCALTZ as a timezone field set to None.
                    timezone=None,
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.INTERVAL_YEAR_MONTH):
                recap_type = BytesType(
                    logical="build.recap.Interval",
                    bytes_=12,
                    signed=True,
                    unit="month",
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.INTERVAL_DAY_TIME):
                recap_type = BytesType(
                    logical="build.recap.Interval",
                    bytes_=12,
                    signed=True,
                    unit="second",
                    **extra_attrs,
                )
            case HMapType(key_type=key_type, value_type=value_type):
                recap_type = MapType(
                    keys=self._parse_schema(key_type),
                    values=self._parse_schema(value_type),
                    **extra_attrs,
                )
            case HListType(element_type=element_type):
                recap_type = ListType(
                    values=self._parse_schema(element_type),
                    **extra_attrs,
                )
            case HUnionType(types=types):
                # Hive columns are always nullable and default is always null.
                types = [NullType()] + [self._parse_schema(t) for t in htype.types]
                recap_type = UnionType(types=types, default=None, **extra_attrs)
                recap_type = self._flatten_unions(recap_type)
            case HStructType(names=names, types=types):
                recap_type = StructType(
                    fields=[
                        self._parse_schema(HColumn(n, t)) for n, t in zip(names, types)
                    ],
                    **extra_attrs,
                )
            case _:
                raise ValueError(f"Unsupported type: {htype}")

        if not isinstance(recap_type, UnionType) and not isinstance(
            recap_type,
            NullType,
        ):
            recap_type = recap_type.make_nullable()

        return recap_type

    def _flatten_unions(self, recap_type: RecapType) -> RecapType:
        """
        Flattens nested unions into a single union. Useful for unwinding a
        HUnionType since we're wrapping all fields in _parse_schema() in a
        union. Without this, we'd get [NullType, [NullType, IntType],
        [NullType, StrType], ...]. Now we get [NullType, IntType, StrType].
        """

        extra_attrs = recap_type.extra_attrs | {
            "logical": recap_type.logical,
            "alias": recap_type.alias,
            "doc": recap_type.doc,
        }

        if isinstance(recap_type, UnionType):
            # Start with empty set of flattened types
            flattened_types = []

            for union_subtype in recap_type.types:
                # We're not expecting string type aliases here.
                assert isinstance(union_subtype, RecapType)

                # Recursively flatten unions within unions
                flattened = self._flatten_unions(union_subtype)

                if isinstance(flattened, UnionType):
                    # Append types from inner union to the flattened types
                    flattened_types.extend(
                        [t for t in flattened.types if t not in flattened_types]
                    )
                elif flattened not in flattened_types:
                    # Append non-union types directly
                    flattened_types.append(flattened)

            return UnionType(
                types=flattened_types,
                **extra_attrs,
            )
        elif isinstance(recap_type, StructType):
            return StructType(
                fields=[self._flatten_unions(field) for field in recap_type.fields],
                **extra_attrs,
            )
        elif isinstance(recap_type, ListType):
            return ListType(
                values=self._flatten_unions(recap_type.values),
                extra_attrs=extra_attrs,
            )
        elif isinstance(recap_type, MapType):
            return MapType(
                keys=self._flatten_unions(recap_type.keys),
                values=self._flatten_unions(recap_type.values),
                extra_attrs=extra_attrs,
            )
        else:
            # No nested union type to flatten
            return recap_type



from recap.clients import create_client
from recap.settings import RecapSettings
from recap.types import StructType

settings = RecapSettings()


def ls(path: str = "/") -> list[str] | None:
    """
    List the children of a path.

    :param path: Path to list children of. Defaults to root.
    :return: List of children.
    """

    system, *args = _args(path) if path not in ("", "/") else [None]
    if not system:
        return list(settings.systems.keys())
    if system and (url := settings.systems.get(system)):
        with create_client(url.unicode_string()) as client:
            return client.ls(*args)


def schema(path: str) -> StructType | None:
    """
    Get the schema of a path.

    :param path: Path to get schema of.
    :return: Schema of path.
    """

    system, *args = _args(path)
    if system and (url := settings.systems.get(system)):
        with create_client(url.unicode_string()) as client:
            return client.get_schema(*args)


def _args(path: str | None) -> list[str]:
    return path.strip("/").split("/") if path else []



from json import loads

import psycopg2
import pytest
from typer.testing import CliRunner

from recap.cli import app

runner = CliRunner()


class TestCli:
    @classmethod
    def setup_class(cls):
        # Connect to the PostgreSQL database
        cls.connection = psycopg2.connect(
            host="localhost",
            port="5432",
            user="postgres",
            password="password",
            dbname="testdb",
        )

        # Create tables
        cursor = cls.connection.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS test_types (test_integer INTEGER);")
        cls.connection.commit()

    @classmethod
    def teardown_class(cls):
        # Delete tables
        cursor = cls.connection.cursor()
        cursor.execute("DROP TABLE IF EXISTS test_types;")
        cls.connection.commit()

        # Close the connection
        cls.connection.close()

    @pytest.mark.parametrize(
        "args",
        [
            ["ls"],
            ["ls", "/"],
        ],
    )
    def test_ls_root(self, args: list[str]):
        result = runner.invoke(app, args)
        assert result.exit_code == 0
        assert loads(result.stdout) == ["pg"]

    @pytest.mark.parametrize(
        "args",
        [
            ["ls", "pg"],
            ["ls", "/pg"],
        ],
    )
    def test_ls_subpath(self, args: list[str]):
        result = runner.invoke(app, args)
        assert result.exit_code == 0
        assert loads(result.stdout) == ["postgres", "template0", "template1", "testdb"]

    def test_schema(self):
        result = runner.invoke(app, ["schema", "pg/testdb/public/test_types"])
        assert result.exit_code == 0
        assert loads(result.stdout) == {
            "type": "struct",
            "fields": [{"type": "int32", "name": "test_integer", "optional": True}],
        }
